/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.beans.session;

import java.io.IOException;
import java.io.Serializable;

import javax.faces.application.FacesMessage;
import javax.faces.application.FacesMessage.Severity;
import javax.faces.bean.ManagedBean;
import javax.faces.bean.SessionScoped;
import javax.faces.context.FacesContext;

import kieker.analysis.AnalysisController;
import kieker.analysis.exception.AnalysisConfigurationException;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.common.ACManager;
import kieker.webgui.common.Global;
import kieker.webgui.common.exception.AnalysisNotInstantiatedException;
import kieker.webgui.common.exception.AnalysisNotRunningException;
import kieker.webgui.common.exception.ProjectAlreadyStartedException;
import kieker.webgui.common.exception.ProjectStillRunningException;

/**
 * This bean contains the project of the current (session) user for the controller page.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
@ManagedBean
@SessionScoped
public class CurrentControllerBean implements Serializable {

	/**
	 * The serial version UID.
	 */
	private static final long serialVersionUID = 2070646497643630097L;
	/**
	 * This is the log for errors, exceptions etc.
	 */
	private static final Log LOG = LogFactory.getLog(CurrentControllerBean.class);
	/**
	 * This is the name of the stored project. It can be used as an identifier within the FS-Manager
	 */
	private String projectName;
	/**
	 * This is the actual model instance. It is the in-memory-model of the current (session) user.
	 */
	private MIProject project;

	/**
	 * Creates a new instance of this class.
	 */
	public CurrentControllerBean() {
		// No code necessary
	}

	/**
	 * This method sets the project stored within this bean and returns the new page for the navigation.
	 * 
	 * @param newProject
	 *            The project to be stored in this bean.
	 * @param newName
	 *            The name of the project.
	 * @return The name of the page for the cockpit.
	 */
	public String setProject(final MIProject newProject, final String newName) {
		synchronized (this) {
			// Remember the given parameters
			this.project = newProject;
			this.projectName = newName;

			// Now deliver the correct navigation page
			final String navigationPage;
			if (this.project != null) {
				navigationPage = Global.PAGE_ANALYSIS_CONTROLLER;
			} else {
				navigationPage = "";
			}
			return navigationPage;
		}
	}

	/**
	 * This method delivers the project stored in this bean.
	 * 
	 * @return The project for this user.
	 */
	public MIProject getProject() {
		synchronized (this) {
			return this.project;
		}
	}

	/**
	 * This method delivers the project name stored in this bean.
	 * 
	 * @return The project name for this user.
	 */
	public String getProjectName() {
		synchronized (this) {
			return this.projectName;
		}
	}

	/**
	 * This method clears the bean. In other words: The stored project is set to null and it will return the page of the project overview for navigation purposes.
	 * 
	 * @return The name of the page of the project overview.
	 */
	public String clearProject() {
		synchronized (this) {
			this.projectName = null; // NOPMD
			this.project = null; // NOPMD
		}

		return Global.PAGE_PROJECT_OVERVIEW;
	}

	/**
	 * This method starts the current analysis and informs the user about a fail.
	 */
	public void startAnalysis() {
		try {
			ACManager.getInstance().startAnalysisController(this.projectName);
		} catch (final ProjectAlreadyStartedException ex) {
			CurrentControllerBean.LOG.info("The analysis is already running.", ex);
			CurrentControllerBean.showMessage(FacesMessage.SEVERITY_WARN, "The analysis is already running.");
		} catch (final AnalysisNotInstantiatedException ex) {
			CurrentControllerBean.LOG.info("The analysis has not been instantiated yet.", ex);
			CurrentControllerBean.showMessage(FacesMessage.SEVERITY_WARN, "The analysis has not been instantiated yet.");
		}
	}

	/**
	 * This method stops the current analysis and informs the user about a fail.
	 */
	public void stopAnalysis() {
		try {
			ACManager.getInstance().stopAnalysisController(this.projectName);
		} catch (final AnalysisNotRunningException ex) {
			CurrentControllerBean.LOG.info("The analysis has not been started yet.", ex);
			CurrentControllerBean.showMessage(FacesMessage.SEVERITY_WARN, "The analysis has not been started yet.");
		} catch (final InterruptedException ex) {
			CurrentControllerBean.LOG.error("Unknown interrupted exception.", ex);
		}
	}

	/**
	 * This method initializes the current analysis and informs the user about a fail.
	 */
	public void instantiateAnalysis() {
		try {
			ACManager.getInstance().instantiateAnalysisController(this.projectName);
		} catch (final NullPointerException ex) { // NOPMD (Exception is explicitly thrown)
			CurrentControllerBean.LOG.error("An error occurred while instantiating the analysis.", ex);
			CurrentControllerBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occurred while instantiating the analysis.");
		} catch (final AnalysisConfigurationException ex) {
			CurrentControllerBean.LOG.error("An error occurred while instantiating the analysis.", ex);
			CurrentControllerBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occurred while instantiating the analysis.");
		} catch (final IOException ex) {
			CurrentControllerBean.LOG.error("An error occurred while instantiating the analysis.", ex);
			CurrentControllerBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occurred while instantiating the analysis.");
		} catch (final ProjectStillRunningException ex) {
			CurrentControllerBean.LOG.info("The analysis is still running.", ex);
			CurrentControllerBean.showMessage(FacesMessage.SEVERITY_ERROR, "The analysis is still running.");
		}
	}

	/**
	 * This method cleans the current analysis instance.
	 */
	public void cleanAnalysis() {
		ACManager.getInstance().cleanAnalysisController(this.projectName);
	}

	/**
	 * Checks whether the analysis is currently running.
	 * 
	 * @return true if and only if the analysis is running.
	 */
	public boolean isAnalysisRunning() {
		return ACManager.getInstance().getAnalysisControllerState(this.projectName) == AnalysisController.STATE.RUNNING;
	}

	/**
	 * Checks whether the analysis is currently in the ready state.
	 * 
	 * @return true if and only if the analysis is ready to be started.
	 */
	public boolean isAnalysisReady() {
		return ACManager.getInstance().getAnalysisControllerState(this.projectName) == AnalysisController.STATE.READY;
	}

	/**
	 * Checks whether the analysis is not available.
	 * 
	 * @return true if and only if the analysis is <b>not</b> available.
	 */
	public boolean isAnalysisNotAvailable() {
		return ACManager.getInstance().getAnalysisControllerState(this.projectName) == null;
	}

	/**
	 * Checks whether the analysis is currently terminated.
	 * 
	 * @return true if and only if the analysis has been terminated.
	 */
	public boolean isAnalysisTerminated() {
		return ACManager.getInstance().getAnalysisControllerState(this.projectName) == AnalysisController.STATE.TERMINATED;
	}

	/**
	 * Checks whether the analysis is currently in the failed state.
	 * 
	 * @return true if and only if the analysis has failed.
	 */
	public boolean isAnalysisFailed() {
		return ACManager.getInstance().getAnalysisControllerState(this.projectName) == AnalysisController.STATE.FAILED;
	}

	/**
	 * This method shows the current user a message by using the growl-component of PrimeFaces.
	 * 
	 * @param severity
	 *            The severity of the message.
	 * @param msg
	 *            The message itself.
	 */
	private static void showMessage(final Severity severity, final String msg) {
		FacesContext.getCurrentInstance().addMessage(null, new FacesMessage(severity, "", msg));
	}
}
